/************************************************************************/
/*  Routines to select best input stream.				*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2000 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char sccsid[] = "$Id: find_best_stream.c,v 1.3 2000/10/26 13:39:49 doug Exp $ ";
#endif

#include <stdio.h>

#include "qlib2.h"

#include "params.h"
#include "st_info.h"
#include "procs.h"
#include "externals.h"

/************************************************************************/
/*  find_best_stream:							*/
/*	Find the best input stream from which to read.			*/
/************************************************************************/
ST_INFO *find_best_stream 
   (ST_INFO	*in_p,		/* ptr to current input stream.		*/
    ST_INFO	*out_p,		/* ptr to output stream.		*/
    INT_TIME	*stime,		/* ptr to desired start time, or NULL.	*/
    INT_TIME	*etime,		/* ptr to desired end time, or NULL.	*/
    int		bias)		/* bias flag for best stream selection.*/
{
    ST_INFO	*st_p;
    ST_INFO	*best_p = NULL;
    INT_TIME	xbegtime, xendtime;
    double	tdif;

    /*	If either we have a start time OR we have no current stream,	*/
    /*	then position all input streams.				*/
    if (stime != NULL || in_p == NULL) 
	for (st_p = st_head_in.next; st_p != NULL; st_p = st_p->next) {
	    /*	If possible, use header time for a fast check to	*/
	    /*	determine whether desired time interval is in this file.*/
	    if ( (! ignore_time) && (! missing_time(st_p->begtime)) && 
		(! missing_time(st_p->endtime)) && 
		(stime != NULL | etime != NULL) ) {
		xbegtime = add_time (st_p->begtime, -1*MAX_SEQ_TIME_OVERLAP, 0);
		xendtime = add_time (st_p->endtime, MAX_SEQ_TIME_OVERLAP, 0);
		if ((stime != NULL && tdiff(xendtime,*stime) < 0) ||
		    (etime != NULL && tdiff(xbegtime,*etime) > 0)) {
		    st_p = close_stream(st_p);
		    continue;
		}
	    }
	    if (position_stream (st_p, stime, etime) == EOF) 
		st_p = close_stream(st_p);
	}

    if (in_p != NULL) {
	/*  Get the next block of the current input stream.		*/
	if (position_stream (in_p, stime, etime) == EOF) {
	    /* EOF -- close it and unset current input stream.		*/
	    in_p = close_stream(in_p);
	    in_p = NULL;
	}
    }

    /*	If either we have a SCAN_BIAS OR we have no current stream,	*/
    /*	then position all input streams.				*/

    if (bias == SCAN_BIAS || in_p == NULL) 
	for (st_p = st_head_in.next; st_p != NULL; st_p = st_p->next)
	    if (position_stream (st_p, stime, etime) == EOF) 
		st_p = close_stream(st_p);

    /*	Since the logs are not completely sequential:			*/
    /*	a.  In the worst case, we will have to look at EVERY file	*/
    /*	    EACH time, and always take the EARLIEST block that falls	*/
    /*	    within our time window.					*/
    /*	b.  We must read ALL files to completion.			*/

    /*	If the bias is SCAN_BIAS, this means that we should scan	*/
    /*	ALL input streams EVERY TIME in order to get the most		*/
    /*	"sequential" interleave of data from the input streams.		*/

    /*	If the bias is STAY_BIAS, this means that we should stay	*/
    /*	with the current input file until it ends before trying another	*/
    /*	input file.  Given the basic nature of log files, this is	*/
    /*	probably adequate.						*/

    if ( (in_p != NULL) && (bias == SAME_BIAS) ) return(in_p);
	
    /*	If we have a start time, look for the "best" stream that	*/
    /*	contains the start time.					*/
    /*	If no block intersects the start time, take the earliest block.	*/

    for (st_p = st_head_in.next; st_p != NULL; st_p = st_p->next) {
	/* Take stream with the earliest time.			*/
	if (best_p == NULL) { best_p = st_p; continue; }
	tdif = tdiff(st_p->cur_hdr->begtime,best_p->cur_hdr->begtime);
	if (tdif <= 0) best_p = st_p;
    }
    return (best_p);
}

