/************************************************************************/
/*  Routines to process current input record.				*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2000 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

/*
96/11/01 DSN	Removed usec99 from time for blockettes.
*/

#ifndef lint
static char sccsid[] = "$Id: process_block.c,v 1.12 2009/05/05 02:16:40 doug Exp $ ";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <math.h>

#include "qlib2.h"

#include "params.h"
#include "st_info.h"
#include "procs.h"
#include "externals.h"

#define	    SIGN(i)	( (i>=0) ? 1 : -1 )

#define	    OUTPUT_BLKSIZE	    4096
#define	    OUT_FRAMES_PER_BLOCK    ((OUTPUT_BLKSIZE-FRAMESIZE)/FRAMESIZE)

double	dsamples_in_time();
void	make_out_sum_hdr();
void	make_out_cur_hdr();
char *q_clock_model(char clock_model);
char *q_clock_status(char *status, char clock_model);

/************************************************************************/
/*  process_block:							*/
/*	Process input block for output.					*/
/************************************************************************/
void process_block
   (ST_INFO	*st_p)		/* ptr to input stream structure.	*/
{
    INT_TIME	    exptime, *p_last_time;
    ST_INFO	    *out = st_head_out.next;
    int		    begin_trim_pts;
    int		    end_trim_pts;
    int		    ntrim;
    int		    bdif;
    int		    edif;
    int		    ok_to_merge;
    int		    x0, xm1, d0;

    if (my_wordorder < 0) my_wordorder = get_my_wordorder();

reprocess:
    /*	Ensure that all headers for the output stream are allocated	*/
    /*	and initialized.						*/
    if (out->sum_hdr == NULL) make_out_sum_hdr(out, st_p);
    
    if (out->cur_hdr == NULL) make_out_cur_hdr(out);

    /*	Perform sanity checks on the input stream.			*/
    if (! same_type_data_stream (st_p->cur_hdr, out->sum_hdr)) {
	fprintf (info, "input block for different stream than output stream\n");
	fprintf (info, "output: sncl=%s.%s.%s.%s,  input:  sncl=%s.%s.%s.%s\n",
		 out->sum_hdr->station_id, out->sum_hdr->network_id, out->sum_hdr->channel_id, out->sum_hdr->location_id,
		 st_p->cur_hdr->station_id, st_p->cur_hdr->network_id, st_p->cur_hdr->channel_id, st_p->cur_hdr->location_id);

	exit(1);
    }

    bdif = 0;
    append_to_output (st_p, bdif);

    /*	Update input header structures.					*/
    if (st_p->prev_hdr != NULL) free_data_hdr (st_p->prev_hdr);
    st_p->prev_hdr = st_p->cur_hdr;
    st_p->cur_hdr = NULL;
    return;
}

/************************************************************************/
/*  pad_out:								*/
/*	Pad the output buffer (if not empty).				*/
/************************************************************************/
int pad_out
   (ST_INFO	*out)		/* ptr to output stream structure.	*/
{
    return (0);
}


/************************************************************************/
/*  flush_output:							*/
/*	Flush any buffered output.					*/
/************************************************************************/
void flush_output
   (ST_INFO	*out)		/* ptr to output stream structure.	*/
{
    IOB		    *iob = out->iob;
    DATA_HDR	    *ch = out->cur_hdr;

    pad_out(out);
	if debug(DEBUG_OUT) {
	    fprintf (info, "out: %s.%s.%s.%s, block %d, %d points (%s to ",
		     ch->station_id, ch->network_id, ch->channel_id, ch->location_id,  
		     ch->seq_no, ch->num_samples,
		     time_to_str(ch->begtime,date_fmt));
	    fprintf (info, "%s)\n", time_to_str(ch->endtime,date_fmt));
	}
}

/************************************************************************/
/*  append_to_output:							*/
/*	Append input to output.						*/
/************************************************************************/
void append_to_output
   (ST_INFO	*st_p,		/* ptr to input stream structure.	*/
    int		slew)		/* slew between current and prev record.*/
{
    ST_INFO	*out = st_head_out.next;
    DATA_HDR	*oh = out->cur_hdr;
    DATA_HDR	*ih = st_p->cur_hdr;
    DATA_HDR	*sh = out->sum_hdr;
    int		i_offset = ih->first_data;
    int		i_len = ih->num_data_frames * FRAMESIZE;
    int		o_offset = oh->first_data +
			   (oh->num_data_frames * FRAMESIZE);
    STORE_LOG	*sp;
    STORE_EVENT	*se;
    char	*comment_type;
    INT_TIME	dtime;
    char	*pstr, *p;
    int		i, nc;
    char	sncl[80];

    sprintf (sncl, "%s.%s.%s.%s", ih->station_id, ih->network_id, ih->channel_id, ih->location_id);
    switch (ih->data_type) {
	BLOCKETTE_200 *b200;
	BLOCKETTE_201 *b201;
	BLOCKETTE_300 *b300;
	BLOCKETTE_310 *b310;
	BLOCKETTE_320 *b320;
	BLOCKETTE_390 *b390;
	BLOCKETTE_395 *b395;
	BLOCKETTE_400 *b400;
	BLOCKETTE_405 *b405;
	BLOCKETTE_500 *b500;
	BLOCKETTE_1000 *b1000;
	BLOCKETTE_1001 *b1001;
	BLOCKETTE_2000 *b2000;
    case    STYPE_EVENT:
	/*  Output event info.	*/
	se = (STORE_EVENT *) st_p->iob->buf;
	dtime = decode_time_drm (se->da_time,ih->hdr_wordorder);
	pstr = time_to_str(dtime, date_fmt);
#ifdef	STATION_INFO
	fprintf (out->iob->fp, "%-16.16s", sncl);
#endif
	fprintf (out->iob->fp, "%s %s detection_type= %d ",
		 pstr, "(evt)", se->detection_type);
	fprintf (out->iob->fp, "motion_quality= %x peak_amp= %d period*100= %d background_amp= %d\n",
		 se->motion_qual, se->peak_amp, se->period100, se->back_amp);
	break;
      case    STYPE_LOG:
	/*  Output log info.	*/
	sp = (STORE_LOG *) st_p->iob->buf;
	comment_type =  (sp->comment_type == 0) ? "(con)" :
			(sp->comment_type == CMT_CLOCK_CORRECTION) ? "(clk)" :
			(sp->comment_type == CMT_TIME_GAP) ? "(gap)" : 
			"(unk)";
	pstr = time_to_str(ih->begtime, date_fmt);
#ifdef	STATION_INFO
	fprintf (out->iob->fp, "%-16.16s", sncl);
#endif
	fprintf (out->iob->fp, "%s %s %s\n", pstr, comment_type, sp->comment);
	break;
      case    SDR_TYPE:
	/* dump out blockettes */
	if (ih->pblockettes != NULL) {
	    int next;
	    int l;
	    int type;
	    BS *bs;
	    BLOCKETTE_HDR *bh;
	    char detector_name[25];
	    bs = ih->pblockettes;
/*::
	    if ((bs = find_blockette(ih,1001))) {
		bh = (BLOCKETTE_HDR *)(bs->pb);
	    }
::*/
	    do {
		char exception_str[16+1];
		char clock_model_str[32+1];
		char clock_status_str[128+1];
		BLOCKETTE_HDR *bh = (BLOCKETTE_HDR *)(bs->pb);
		/* Convert wordorder of blockette to machine wordorder.  */
		if (bs->wordorder != my_wordorder) {
		    if (swab_blockette (bs->type, bs->pb, bs->len) == 0) {
			bs->wordorder = my_wordorder;
		    }
		    else {
			fprintf (stderr, "Error: Unable to swab blockette %d\n", bs->type);
		    }
		}
		switch (type=bs->type) {
		  case 200:		/* Quanterra Threshold event detector.	*/
#ifdef	STATION_INFO
		    fprintf (out->iob->fp, "%-16.16s", sncl);
#endif
		    fprintf (out->iob->fp, "%s %s ",
			     time_to_str(decode_time_sdr(((BLOCKETTE_200 *)bh)->time,
							  bs->wordorder),
					  date_fmt), "(thr)");
		    fprintf (out->iob->fp, "amp= %.2f limit= %.2f",
			     (((BLOCKETTE_200 *)bh)->signal_amplitude),
			     (((BLOCKETTE_200 *)bh)->background_estimate));
		    /* Print detector name if available.    */
		    if (bs->len > BLOCKETTE_200_STD_SIZE) {
			charncpy(detector_name,((BLOCKETTE_200 *)bh)->detector_name,24);
			trim(detector_name);
			fprintf (out->iob->fp, " det_name= %s", detector_name);
		    }
		    fprintf (out->iob->fp, "\n");
		    oh->num_samples++;
		    break;
		  case 201:		/* Murdock Hutt event detector.		*/
#ifdef	STATION_INFO
		    fprintf (out->iob->fp, "%-16.16s", sncl);
#endif
		    fprintf (out->iob->fp, "%s %s ",
			     time_to_str(decode_time_sdr(((BLOCKETTE_201 *)bh)->time,
							  bs->wordorder), 
					  date_fmt), "(mhh)");
		    fprintf (out->iob->fp, "det= %x pick= %d loopback= %d snr= %0d%d%d%d%d peak_amp= %.0f period= %.2f background_amp= %.0f ",
			     (((BLOCKETTE_201 *)bh)->detection_flags),
			     (((BLOCKETTE_201 *)bh)->pick_algorithm),
			     (((BLOCKETTE_201 *)bh)->loopback_value),
			     (((BLOCKETTE_201 *)bh)->signal_to_noise[0]),
			     (((BLOCKETTE_201 *)bh)->signal_to_noise[1]),
			     (((BLOCKETTE_201 *)bh)->signal_to_noise[2]),
			     (((BLOCKETTE_201 *)bh)->signal_to_noise[3]),
			     (((BLOCKETTE_201 *)bh)->signal_to_noise[4]),
			     (((BLOCKETTE_201 *)bh)->signal_amplitude),
			     (((BLOCKETTE_201 *)bh)->signal_period),
			     (((BLOCKETTE_201 *)bh)->background_estimate));
		    if (bs->len > BLOCKETTE_201_STD_SIZE) {
			charncpy(detector_name,((BLOCKETTE_201 *)bh)->detector_name,24);
			trim(detector_name);
			fprintf (out->iob->fp, " det_name= %s", detector_name);
		    }
		    fprintf (out->iob->fp, "\n");
		    oh->num_samples++;
		    break;
	          case 300:		/* Step Calibration blockette.		*/
#ifdef	STATION_INFO
		    fprintf (out->iob->fp, "%-16.16s", sncl);
#endif
		    fprintf (out->iob->fp, "%s %s ",
			     time_to_str(decode_time_sdr(((BLOCKETTE_300 *)bh)->time,
							  bs->wordorder), 
					  date_fmt), "(calstep)");
		    fprintf (out->iob->fp, "nstep= %d flags= %x step_dur= %d int_dur= %d cal_amp= %.4f ",
			     (((BLOCKETTE_300 *)bh)->num_step_calibrations),
			     (((BLOCKETTE_300 *)bh)->calibration_flags),
			     (((BLOCKETTE_300 *)bh)->step_duration),
			     (((BLOCKETTE_300 *)bh)->interval_duration),
			     (((BLOCKETTE_300 *)bh)->calibration_amplitude));
		    if (bs->len > BLOCKETTE_300_STD_SIZE) {
			fprintf (out->iob->fp, "ref_amp= %.4f coupling= %-.12s rolloff= %-.12s",
				 (((BLOCKETTE_300 *)bh)->reference_amplitude),
				 (((BLOCKETTE_300 *)bh)->coupling),
				 (((BLOCKETTE_300 *)bh)->rolloff));
		    }
		    fprintf (out->iob->fp, "\n");
		    oh->num_samples++;
		    break;
		  case 310:		/* Sine Calibration blockette.		*/
#ifdef	STATION_INFO
		    fprintf (out->iob->fp, "%-16.16s", sncl);
#endif
		    fprintf (out->iob->fp, "%s %s ",
			     time_to_str(decode_time_sdr(((BLOCKETTE_310 *)bh)->time,
							  bs->wordorder), 
					  date_fmt), "(calsine)");
		    fprintf (out->iob->fp, "flags= %x cal_dur= %d cal_period= %.4f cal_amp= %.4f ",
			     (((BLOCKETTE_310 *)bh)->calibration_flags),
			     (((BLOCKETTE_310 *)bh)->calibration_duration),
			     (((BLOCKETTE_310 *)bh)->calibration_period),
			     (((BLOCKETTE_310 *)bh)->calibration_amplitude));
		    if (bs->len > BLOCKETTE_310_STD_SIZE) {
			fprintf (out->iob->fp, "ref_amp= %.4f coupling= %-.12s rolloff= %-.12s",
				 (((BLOCKETTE_310 *)bh)->reference_amplitude),
				 (((BLOCKETTE_310 *)bh)->coupling),
				 (((BLOCKETTE_310 *)bh)->rolloff));
		    }
		    fprintf (out->iob->fp, "\n");
		    oh->num_samples++;
		    break;
		  case 320:		/* Pseudo-random Calibration blockette.	*/
#ifdef	STATION_INFO
		    fprintf (out->iob->fp, "%-16.16s", sncl);
#endif
		    fprintf (out->iob->fp, "%s %s ",
			     time_to_str(decode_time_sdr(((BLOCKETTE_320 *)bh)->time,
							  bs->wordorder), 
					  date_fmt), "(calrand)");
		    fprintf (out->iob->fp, "flags= %x cal_dur= %d cal_amp= %.4f ",
			     (((BLOCKETTE_320 *)bh)->calibration_flags),
			     (((BLOCKETTE_320 *)bh)->calibration_duration),
			     (((BLOCKETTE_320 *)bh)->calibration_amplitude));
		    if (bs->len > BLOCKETTE_320_STD_SIZE) {
			fprintf (out->iob->fp, "ref_amp= %.4f coupling= %-.12s rolloff= %-.12s noise= %-.8s",
				 (((BLOCKETTE_320 *)bh)->reference_amplitude),
				 (((BLOCKETTE_320 *)bh)->coupling),
				 (((BLOCKETTE_320 *)bh)->rolloff),
				 (((BLOCKETTE_320 *)bh)->noise_type));
		    }
		    fprintf (out->iob->fp, "\n");
		    oh->num_samples++;
		    break;
		  case 390:		/* Generic Calibration blockette.	*/
#ifdef	STATION_INFO
		    fprintf (out->iob->fp, "%-16.16s", sncl);
#endif
		    fprintf (out->iob->fp, "%s %s ",
			     time_to_str(decode_time_sdr(((BLOCKETTE_390 *)bh)->time,
							  bs->wordorder), 
					  date_fmt), "(calgen)");
		    fprintf (out->iob->fp, "flags= %x cal_dur= %d cal_amp= %.4f ",
			     (((BLOCKETTE_390 *)bh)->calibration_flags),
			     (((BLOCKETTE_390 *)bh)->calibration_duration),
			     (((BLOCKETTE_390 *)bh)->calibration_amplitude));
		    fprintf (out->iob->fp, "\n");
		    oh->num_samples++;
		    break;
		  case 395:		/* Calibration Abort blockette.		*/
#ifdef	STATION_INFO
		    fprintf (out->iob->fp, "%-16.16s", sncl);
#endif
		    fprintf (out->iob->fp, "%s %s ",
			     time_to_str(decode_time_sdr(((BLOCKETTE_395 *)bh)->time,
							  bs->wordorder), 
					  date_fmt), "(calabort)");
		    fprintf (out->iob->fp, "\n");
		    oh->num_samples++;
		    break;
		  case 400:		/* Beam blockette.			*/
		    fprintf (info, "blockette %d\n",type);
		    break;
		  case 405:		/* Beam Delay blockette.		*/
		    fprintf (info, "blockette %d\n",type);
		    break;
		  case 500:		/* Timing blockette.			*/
		    b500 = (BLOCKETTE_500 *)bh;
#ifdef	STATION_INFO
		    fprintf (out->iob->fp, "%-16.16s", sncl);
#endif
		    fprintf (out->iob->fp, "%s %s ",
			     time_to_str(decode_time_sdr(b500->time, my_wordorder), 
					  date_fmt), "(clk)");
		    charncpy (exception_str,b500->exception_type, 16);
		    charncpy (clock_model_str,b500->clock_model, 32);
		    charncpy (clock_status_str,b500->clock_status, 128);
		    trim(exception_str);
		    trim(clock_model_str);
		    trim(clock_status_str);
		    fprintf (out->iob->fp, "%s timemark, count= %d ",exception_str, 
			     b500->count);
		    fprintf (out->iob->fp, "quality= %d vco_corr= %f ", 
			     b500->reception_quality, b500->vco_correction);
		    fprintf (out->iob->fp, "clock_model= %s clock_status= %s\n",
			     clock_model_str, clock_status_str);
		    oh->num_samples++;
		    break;
		  case 1000:	/* Data format blockette.		*/
		    break;
		  case 1001:	/* Data extention blockette.		*/
		    break;
		  case 2000:	/* Opaque data blockette.		*/
		    /* Output the data portion of the record.		*/
		    fwrite ((char *)bh+((BLOCKETTE_2000 *)bh)->data_offset,
			    ((BLOCKETTE_2000 *)bh)->blockette_len-((BLOCKETTE_2000 *)bh)->data_offset,
			    1, out->iob->fp);
		    break;
		  default:
		    fprintf (info,   "unknown blockette %d, skipping\n", bh->type);
		    fprintf (stderr, "unknown blockette %d, skipping\n", bh->type);
		}
		bs = bs->next;
	    } while (bs != (BS *)NULL);
	}
	break;
      default:
	fprintf (info, "invalid data_type %d -- should never get here\n",
		 ih->data_type);
	exit(1);
	break;
    }

    /*  Process any text messages in this block.				*/
    for (p=st_p->iob->buf+ih->first_data, nc=0, i=0; i<ih->num_samples; i++) {
	if (p[i] == '\r') continue;
	if (nc++ == 0) {
#ifdef	STATION_INFO
	    fprintf (out->iob->fp, "%-16.16s", sncl);
#endif
	    fprintf(out->iob->fp,"%s ",time_to_str(ih->begtime, date_fmt));
	}
	fputc(p[i],out->iob->fp);
	if (p[i] == '\n') {
	    oh->num_samples++;
	    nc = 0;
	}
    }
    /*	Update the input buffer pointers to point past this the input	*/
    /*	data block.							*/
    st_p->iob->avail += st_p->blksize;

    /*	Update output header info.					*/
    if (oh->num_samples == 0) {
	memcpy ((void *)&oh->begtime, (void *)&ih->begtime, sizeof oh->begtime);
    }
    memcpy ((void *)&oh->endtime, (void *)&ih->endtime, sizeof(INT_TIME));

    /*	Update summary header info.					*/
    if (sh->num_samples == 0) {
	memcpy ((void *)&sh->begtime, (void *)&ih->begtime, sizeof sh->begtime);
    }
    ++sh->num_samples;
    memcpy ((void *)&sh->endtime, (void *)&ih->endtime, sizeof sh->endtime);
}

/************************************************************************/
/*  same_type_data_stream:						*/
/*	Boolean function to compare stream characteristics.		*/
/************************************************************************/
int same_type_data_stream
   (DATA_HDR	*ih,		/* ptr to input DATA_HDR.		*/
    DATA_HDR	*oh)		/* ptr to output DATA_HDR.		*/
{
    if ((strcmp(ih->station_id, oh->station_id) != 0) ||
	(strcmp(ih->channel_id, oh->channel_id) != 0) ||
	(strcmp(ih->location_id, oh->location_id) != 0) 
/*:: ||
	(ih->data_type != oh->data_type) || 
	(ih->sample_rate != oh->sample_rate) ||
	(ih->sample_rate_mult != oh->sample_rate_mult)
/::*/
    ) return (0);
    else return (1);
}

/************************************************************************/
/*  make_out_sum_hdr:							*/
/*	Alloc and init summary output header struct.			*/
/************************************************************************/
void make_out_sum_hdr
   (ST_INFO	*out,		/* ptr to output stream structure.	*/
    ST_INFO	*st_p)		/* ptr to input stream structure.	*/
{
    /*	Make and fill in summary header if we don't have one yet.	*/
    if (out->sum_hdr == NULL) {
	out->sum_hdr = new_data_hdr();
	/*  Copy info from current input stream to summary header.	*/
	strcpy (out->sum_hdr->station_id, st_p->cur_hdr->station_id);
	strcpy (out->sum_hdr->network_id, st_p->cur_hdr->network_id);
	strcpy (out->sum_hdr->channel_id, st_p->cur_hdr->channel_id);
	strcpy (out->sum_hdr->location_id, st_p->cur_hdr->location_id);
	out->sum_hdr->sample_rate = st_p->cur_hdr->sample_rate;
	out->sum_hdr->sample_rate_mult = st_p->cur_hdr->sample_rate_mult;
	/*  Fill in stream header structure also from summary header.	*/
	strcpy (out->station, out->sum_hdr->station_id);
	strcpy (out->channel, out->sum_hdr->channel_id);
	out->rate = out->sum_hdr->sample_rate;
	out->rate_mult = out->sum_hdr->sample_rate_mult;
	out->hdr_type = SDR_HDR_TYPE;
	out->blksize = OUTPUT_BLKSIZE;
	out->data_type = st_p->data_type;
	out->sum_hdr->data_type = out->data_type;
    }
}

/************************************************************************/
/*  make_out_cur_hdr:							*/
/*	Alloc and init current output header struct.			*/
/************************************************************************/
void make_out_cur_hdr
   (ST_INFO	*out)		/* ptr to output stream structure.	*/
{
    /*	Make and fill in current header if we don't have one yet.	*/
    if (out->cur_hdr == NULL) {
	out->cur_hdr = new_data_hdr();
	memset ((void *)out->cur_hdr, 0, sizeof(DATA_HDR));
	strcpy (out->cur_hdr->station_id, out->sum_hdr->station_id);
	strcpy (out->cur_hdr->network_id, out->sum_hdr->network_id);
	strcpy (out->cur_hdr->channel_id, out->sum_hdr->channel_id);
	strcpy (out->cur_hdr->location_id, out->sum_hdr->location_id);
	out->cur_hdr->sample_rate = out->sum_hdr->sample_rate;
	out->cur_hdr->sample_rate_mult = out->sum_hdr->sample_rate_mult;
	out->cur_hdr->first_data = 64;
	out->cur_hdr->seq_no = (out->prev_hdr == NULL) ? 1 : 
	    out->prev_hdr->seq_no + 1;
	out->cur_hdr->data_type = out->sum_hdr->data_type;
    }
}


