/************************************************************************/
/*  qlog -	merge multiple quanterra log streams into single	*/
/*		continuous output data stream.				*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2000 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char sccsid[] = "$Id: qlog.c,v 1.8 2007/05/09 23:48:27 doug Exp $ ";
#endif

#define	START_BIAS		SCAN_BIAS
#define	RUN_BIAS		SCAN_BIAS
#define	    EXTERN

#include <stdio.h>

#include "version.h"

char *syntax[] = {
"%s version " VERSION,
"%s  [-o output_file | -n] [-f date] [-t date | -s interval] [-D j | d]",
"    [-i time] [-c] [-d n] [-h] [input_file_1 ... input_file_n]",
"    where:",
"	-o outfile  Name of output file.",
"	-n	    No data output.  Equivalent to -o /dev/null.",
"	-f date	    From date - ignore data before this date.",
"	-t date	    To date - ignore data from this date on.",
"		    Date can be in formats:",
"			[19]yy/mm/dd/hh:mm:ss.ffff",
"			[19]yy/mm/dd.hh:mm:ss.ffff",
"			[19]yy/mm/dd,hh:mm:ss.ffff",
"			[19]yy.ddd.hh:mm:ss.ffff",
"			[19]yy.ddd,hh:mm:ss.ffff",
"			[19]yy,ddd,hh:mm:ss.ffff",
"	-s interval span interval.  Alternate way of specifying end time.",
"		    Interval can be an integer followed immediately by",
"		    s, m, h, or d for seconds, minutes, hours, or days.",
"	-D j | m    Display dates in julian or month and day format.",
"	-i time	    Ignore time span info in header.",
"	-c	    Write index consisting of station and channel to stdout.",
"	-d n	    Debug output (OR of any of the following values:)",
"			2 = time info",
"			4 = block info",
"			8 = block headers",
"			16 = flags",
"			32 = input stream",
"			64 = output stream",
"	-h	    Help - prints syntax message.",
"	input_file(s)",
"		    Input file(s) containing quanterra data.",
"		    If no input file is specified, input is read from stdin.",
NULL };

/************************************************************************/

#include "qlib2.h"

#include "params.h"
#include "st_info.h"
#include "procs.h"
#include "externals.h"

main ( int argc, char **argv)
{
    
    ST_INFO	*st_p;		/* ptr to input stream structure.	*/
    ST_INFO	*out_p;		/* ptr to output stream structure.	*/
    ST_INFO	*prev_stream;	/* ptr to prev input stream structure.	*/
    int		i;
    int		bias, start_bias, run_bias;
    INT_TIME	*stime;
    INT_TIME	*etime;

    /*	Initialization.							*/
    outfile = NULL;		/* empty initial output directory.	*/
    info = stdout;
    date_fmt = JULIAN_FMT;
    run_bias = RUN_BIAS;
    if (init_qlib2(1) != 1) {
	fprintf (stderr, "Error: initializing qlib2\n");
	exit(1);
    }
    sec_per_year(1970);		/* force leapsecond initialization.	*/

    debug_option = 0;

    parse_cmdline (&argc, &argv);

    if (debug(DEBUG_TIME)) {
	if (start_flag)
	    fprintf (info, "interval start time = %s\n", time_to_str(start_time,date_fmt));
	if (end_flag)
	    fprintf (info, "interval stop time = %s\n", time_to_str(end_time,date_fmt));
    }

    /*	Determine the block size and format of each input stream.	*/
    for (st_p = st_head_in.next; st_p != NULL; st_p = st_p->next) {
	get_stream_info (st_p);
    }

    /*  If index is desired, return info from first available file.	*/
    if (make_index) {
	if ((st_p = st_head_in.next) != NULL) {
	    fprintf (stdout, "%-5.5s %-3.3s %-2.2s %-2.2s %4d\n", 
		     st_p->station, st_p->channel, 
		     (st_p->network[0]) ? st_p->network : "__",
		     (st_p->location[0]) ? st_p->location : "__",
		     st_p->blksize);
	    exit(0);
	}
	else exit(1);	/* No info for index is an ERROR.		*/
    }

    /*	Get the first block for each stream.				*/
    for (st_p = st_head_in.next; st_p != NULL; st_p = st_p->next) {
	get_block_info (st_p);
    }

    /*	Initial step 0:

	If we don't have a starting time, take the file with the earliest
	starting time.  If we have multiple files with same starting point,
	pick the first one with the LARGEST blocksize.

	If we have a start time, position each stream to the first block such
	that the end of the current block does not preceed the start time.

	Select the (first of equal) streams that have the largest (?) block.
    */
    /*	Induction step i+1:

	If we have an empty output block, pick the best input file.
	If we have multiple files with the same starting point, pick the
	first one with the LARGEST blocksize.

	If we have a non-empty output block, pick the best input file.
	If we have multiple files with the same starting point, pick the
	first one with the SMALLEST blocksize.

	If the input block will not fit into the output block, purge the
	output block and start the induction step over again.
    */

    start_bias = START_BIAS;
    bias = start_bias;

    /*	Initialize search parameters.					*/
    st_p = NULL;
    out_p = st_head_out.next;
    stime = (start_flag) ? &start_time : NULL;
    etime = (end_flag) ? &end_time : NULL;

    /*	Loop reading input and creating output.				*/
    /*	Now find the "best" stream to use.				*/
    /*	By definition, all streams have current block info.		*/
    prev_stream = st_p;
    while ( (st_p = find_best_stream (st_p, out_p, stime, etime, bias)) != NULL ) {
	/*	Process the data in the current block.			*/
	if (prev_stream != st_p) {
	    if (debug(DEBUG_STREAM)) fprintf (info, "time %s using stream %s\n",
		time_to_str(st_p->cur_hdr->begtime,date_fmt), st_p->filename);
	    prev_stream = st_p;
	}
	prev_stream = st_p;
	process_block (st_p);
/*	stime = NULL; */
	bias = run_bias;
    }
    flush_output(out_p);
    fclose(out_p->iob->fp);
    print_stream_summary(out_p);
    return(0);
}

/************************************************************************/
/*  print_stream_summary:						*/
/*	Print stream summary info.					*/
/************************************************************************/
void print_stream_summary
   (ST_INFO	*out_p)		/* ptr to stream structure.		*/
{
    int		seconds, usecs;
    INT_TIME	endtime;

    if (out_p->sum_hdr != NULL) {
    /*	Output summary information for the output stream.		*/
    time_interval2 (1, out_p->sum_hdr->sample_rate, out_p->sum_hdr->sample_rate_mult,
		    &seconds, &usecs);
    fflush (out_p->iob->fp);
    endtime = add_time (out_p->sum_hdr->endtime, seconds, usecs);
    fprintf (info, "%s.%s.%s.%s text (%s to ",
	     out_p->sum_hdr->station_id, out_p->sum_hdr->network_id, 
	     out_p->sum_hdr->channel_id, out_p->sum_hdr->location_id, 
	     time_to_str(out_p->sum_hdr->begtime,date_fmt));
    fprintf (info, "%s) : %d lines\n", 
	     time_to_str(endtime,date_fmt), out_p->sum_hdr->num_samples);
    }
}
